from __future__ import annotations

from typing import List, Union

from pydantic import BaseModel
from spacy.tokens import Doc, Span, Token


class TokenPolarityOutput(BaseModel):
    """A data class for the polarity output of a span, notably allows for
    plotting the output."""

    class Config:
        arbitrary_types_allowed = True

    polarity: float
    token: Token
    span: Span
    negation: Union[Token, None] = None
    intensifiers: List[Token] = []

    def __repr_str__(self, join_str: str) -> str:
        return join_str.join(
            repr(v) if a is None else f"{a}={v!r}"
            for a, v in [
                ("polarity", round(self.polarity, 3)),
                ("token", self.token),
                ("span", self.span),
            ]
        )

    def __lt__(self, other: object):
        if not isinstance(other, (TokenPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, TokenPolarityOutput):
            other = other.polarity
        return self.polarity < other

    def __gt__(self, other: object):
        if not isinstance(other, (TokenPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, TokenPolarityOutput):
            other = other.polarity
        return self.polarity > other

    def __bool__(self):
        return bool(self.polarity)

    def __eq__(self, other: object):
        if not isinstance(other, (TokenPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, TokenPolarityOutput):
            other = other.polarity
        return self.polarity == other


class SpanPolarityOutput(BaseModel):
    """A data class for the polarity output of a span, notably allows for
    plotting the output."""

    class Config:
        arbitrary_types_allowed = True

    negative: float
    neutral: float
    positive: float
    compound: float
    span: Span
    polarities: List[TokenPolarityOutput]

    def __repr_str__(self, join_str: str) -> str:
        return join_str.join(
            repr(v) if a is None else f"{a}={v!r}"
            for a, v in [
                ("neg", round(self.negative, 3)),
                ("neu", round(self.neutral, 3)),
                ("pos", round(self.positive, 3)),
                ("compound", round(self.compound, 4)),
                ("span", self.span),
            ]
        )

    def __lt__(self, other: object):
        if not isinstance(other, (SpanPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, SpanPolarityOutput):
            other = other.compound
        return self.compound < other

    def __gt__(self, other: object):
        if not isinstance(other, (SpanPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, SpanPolarityOutput):
            other = other.compound
        return self.compound > other

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (SpanPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, SpanPolarityOutput):
            other = other.compound
        return self.compound == other


class DocPolarityOutput(BaseModel):
    """A data class for the polarity output of a doc."""

    class Config:
        arbitrary_types_allowed = True

    negative: float
    neutral: float
    positive: float
    compound: float
    doc: Doc
    polarities: List[SpanPolarityOutput]

    def __repr_str__(self, join_str: str) -> str:
        return join_str.join(
            repr(v) if a is None else f"{a}={v!r}"
            for a, v in [
                ("neg", round(self.negative, 3)),
                ("neu", round(self.neutral, 3)),
                ("pos", round(self.positive, 3)),
                ("compound", round(self.compound, 4)),
            ]
        )

    def __lt__(self, other: object):
        if not isinstance(other, (DocPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, DocPolarityOutput):
            other = other.compound
        return self.compound < other

    def __gt__(self, other: object):
        if not isinstance(other, (DocPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, DocPolarityOutput):
            other = other.compound
        return self.compound > other

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (DocPolarityOutput, float)):
            return NotImplemented
        if isinstance(other, DocPolarityOutput):
            other = other.compound
        return self.compound == other



"""Calculation of various readability metrics."""

from typing import Dict, Iterable

from spacy.language import Language
from spacy.tokens import Doc, Span, Token

from asent.getters import (
    make_doc_polarity_getter,
    make_intensifier_getter,
    make_is_contrastive_conj_getter,
    make_is_negated_getter,
    make_is_negation_getter,
    make_span_polarity_getter,
    make_token_polarity_getter,
    make_valance_getter,
)


class Asent:
    """spaCy v.3.0 component for adding an asent sentiment models to `Doc`
    objects.

    Ading the Token extentions 'valence', 'intensifier', 'is_negation',
    'is_negated', 'is_contrastive_conj' and 'polarity' as well as the
    Span and Doc extension 'polarity'.
    """

    def __init__(
        self,
        nlp: Language,
        lexicon: Dict[str, float],
        name: str = "asent",
        intensifiers: Dict[str, float] = {},
        negations: Iterable[str] = set(),
        contrastive_conjugations: Iterable[str] = set(),
        lowercase: bool = True,
        lemmatize: bool = False,
        force: bool = False,
    ):
        """Initialize component.

        Args:
            nlp (Language): A spaCy language pipeline for which to add the component to
            name (str): The name of the component
            lexicon (Dict[str, float]): The lexicion used to look up the valence scores of a word.
            intensifiers (Dict[str, float]): A dictionary of intensifiers (e.g. {"very": 0.293}). Defaults to {}, indicating no intensifiers is used.
            negations (Iterable[str]): A list of negations (e..g "not"). Defaults to an empty set indicatin no negations will be used.
            contrastive_conjugations (Iterable[str]): A list of contrastive conjugations (e.g. "but"). Defaults to empty set indicating no contrastive conjugations will be used.
            lowercase (bool): Should be text be lowercases before looking up in the lexicons? Defaults to True.
            lemmatize (bool): Should be text be lemmatized before looking up in the lexicons? Defaults to False.
            force (bool): Should existing extensions be overwritten? Defaults to False.
        """
        self.name = name

        if (not Token.has_extension("valence")) or (force is True):
            Token.set_extension(
                "valence",
                getter=make_valance_getter(
                    lexicon=lexicon,
                    lowercase=lowercase,
                    lemmatize=lemmatize,
                ),
                force=force,
            )

        if (not Token.has_extension("intensifier")) or (force is True):
            Token.set_extension(
                "intensifier",
                getter=make_intensifier_getter(
                    intensifiers=intensifiers,
                    lemmatize=lemmatize,
                    lowercase=lowercase,
                ),
                force=force,
            )

        if (not Token.has_extension("is_negation")) or (force is True):
            Token.set_extension(
                "is_negation",
                getter=make_is_negation_getter(
                    negations=negations,
                    lemmatize=lemmatize,
                    lowercase=lowercase,
                ),
                force=force,
            )

        if (not Token.has_extension("is_negated")) or (force is True):
            Token.set_extension(
                "is_negated",
                getter=make_is_negated_getter(),
                force=force,
            )

        if (not Token.has_extension("is_contrastive_conj")) or (force is True):
            Token.set_extension(
                "is_contrastive_conj",
                getter=make_is_contrastive_conj_getter(
                    contrastive_conjugations=contrastive_conjugations,
                    lemmatize=lemmatize,
                    lowercase=lowercase,
                ),
                force=force,
            )

        if (not Token.has_extension("polarity")) or (force is True):
            Token.set_extension(
                "polarity",
                getter=make_token_polarity_getter(
                    valence_getter=None,
                    is_negation_getter=None,
                    intensifier_getter=None,
                ),
                force=force,
            )

        if (not Span.has_extension("polarity")) or (force is True):
            Span.set_extension(
                "polarity",
                getter=make_span_polarity_getter(
                    polarity_getter=None,
                    contrastive_conj_getter=None,
                ),
                force=force,
            )

        if (not Doc.has_extension("polarity")) or (force is True):
            doc_pol_getter = make_doc_polarity_getter(span_polarity_getter=None)
            Doc.set_extension(
                "polarity",
                getter=doc_pol_getter,
                force=force,
            )

    def __call__(self, doc: Doc) -> Doc:
        """Run the pipeline component.

        Args:
            doc (Doc): A spaCy document the component should be applied to.

        Returns:
            Doc: A processed spacy Document.
        """
        return doc


DEFAULT_CONFIG = {
    "lexicon": {},
    "intensifiers": {},
    "negations": set(),
    "contrastive_conj": set(),
    "lowercase": True,
    "lemmatize": False,
    "force": True,
}


@Language.factory("asent_v1", default_config=DEFAULT_CONFIG)
def create_asent_component(
    nlp: Language,
    name: str,
    lexicon: Dict[str, float],
    intensifiers: Dict[str, float],
    negations: Iterable[str],
    contrastive_conj: Iterable[str],
    lowercase: bool,
    lemmatize: bool,
    force: bool,
) -> Asent:
    """Allows a asent sentiment pipe to be added to a spaCy pipe using
    nlp.add_pipe("asent_v1").

    Args:
        nlp (Language): A spaCy language pipeline to add the component to.
        name (str): The name of the component.
        lexicon (Dict[str, float]): The lexicion used to look up the valence scores of a
            word.
        intensifiers (Dict[str, float]): A dictionary of intensifiers (e.g.
            {"very": 0.293}).
        negations (Iterable[str]): A list of negations (e.g. "not").
        contrastive_conj (Iterable[str]): A list of contrastive conjugations (e.g.
            "but").
        lowercase (bool): Should be text be lowercases before looking up in the
            lexicons?
        lemmatize (bool): Should be text be lemmatized before looking up in the
            lexicons?
        force (bool): Should existing extensions be overwritten?

    Returns:
        Asent: A sentiment component.
    """

    return Asent(
        nlp,
        name=name,
        lexicon=lexicon,
        intensifiers=intensifiers,
        negations=negations,
        contrastive_conjugations=contrastive_conj,
        lowercase=lowercase,
        lemmatize=lemmatize,
        force=force,
    )



import pytest

import asent

from .test_getters import nlp_dict  # noqa


@pytest.mark.parametrize(
    "example,lang",
    [
        ("I am not very happy", "en"),
        ("jeg er glad", "da"),
        ("jeg er GLAD", "da"),
        ("jeg er MEGET glad", "da"),
        ("jeg er sur", "da"),
        ("jeg er ikke sur", "da"),
        ("jeg er ikke længere sur", "da"),
        ("jeg er ikke længere særligt sur", "da"),
    ],
)
def test_visualize(example: str, lang: str, nlp_dict):  # noqa
    nlp = nlp_dict[lang]
    nlp.add_pipe("asent_" + lang + "_v1")

    doc = nlp(example)

    # test on docs
    asent.visualize(doc, style="prediction")
    asent.visualize(doc, style="analysis")

    # test on spans
    asent.visualize(doc[:2], style="prediction")
    asent.visualize(doc[:2], style="prediction-no-overlap")
    asent.visualize(doc[:2], style="analysis")

    # error
    with pytest.raises(ValueError):
        asent.visualize(doc[:2], style="invalid")



import json
import os

from ..utils import lexicons


def read_emoji():
    path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..",
        "lexicons",
        "emoji_v1.json",
    )
    with open(path) as f:
        emoji_lexicon = json.load(f)
    return emoji_lexicon


LEXICON = read_emoji()

lexicons.register("emoji_v1", func=LEXICON)



from spacy.language import Language

from ..component import Asent
from ..utils import LEXICON_PATH, components, lexicons, read_lexicon
from .emoji import LEXICON as E_LEXICON

LANG = "no"
LEXICON = read_lexicon(LEXICON_PATH / "no_lexicon_v1.txt")
NEGATIONS = {"ikke", "ik", "ikk", "ik'", "aldrig", "ingen"}
CONTRASTIVE_CONJ = {"dog", "men"}
INTENSIFIERS = {
    "Temmelig ": 0.293,  # bokmål
    "Meget ": 0.293,
    "Mega": 0.293,
    "Litt": 0.293,
    "Ekstremt": 0.293,
    "Totalt": 0.293,
    "Utrolig": 0.293,
    "Veldig ": 0.293,
    "Bestemt": 0.293,
    "Seriøst": 0.293,
    "Absolutt ": 0.293,
    "Helt": 0.293,
    "Betydelig": 0.293,
    "Dypt": 0.293,
    "Djupt": 0.293,
    "Effing": 0.293,
    "Effin": 0.293,
    "Enormt": 0.293,
    "Fremfor alt": 0.293,
    "Spesiell": 0.293,
    "Eksepsjonelt": 0.293,
    "Ytterst": 0.293,
    "Fantastisk": 0.293,
    "Flipping": 0.293,
    "Flippin": 0.293,
    "Frackin": 0.293,
    "Fracking": 0.293,
    "Frickin": 0.293,
    "Fricking": 0.293,
    "Friggin": 0.293,
    "Frigging": 0.293,
    "Fuckin": 0.293,
    "Fucking": 0.293,
    "Fuggin": 0.293,
    "Fugging": 0.293,
    "Hella": 0.293,
    "Høyt": 0.293,
    "Fullt ": 0.293,
    "Høyst": 0.293,
    "Særdeles": 0.293,
    "Intenst": 0.293,
    "Intensivt": 0.293,
    "Sykt": 0.293,
    "Mer": 0.293,
    "Mest": 0.293,
    "Særlig": 0.293,
    "Ganske": 0.293,
    "Bemerkelsesverdig": 0.293,
    "Bare": 0.293,
    "Så": 0.293,
    "Vesentlig": 0.293,
    "Gjennomgående": 0.293,
    "Grundig": 0.293,
    "Svært": 0.293,
    "Uber": 0.293,
    "Uvanlig": 0.293,
    "Mye": 0.293,
    "Maskimalt": 0.293,
    "Nesten": -0.293,
    "Knapt": -0.293,
    "Marginalt": -0.293,
    "Med nød og neppe ": -0.293,
    "Så vidt": -0.293,
    "Omtrent": -0.293,
    "Noenlunde": -0.293,
    "Mindre": -0.293,
    "Lite": -0.293,
    "Iblant": -0.293,
    "Delvis": -0.293,
    "Sjelden": -0.293,
    "Sjelden(t)": -0.293,
    "Nå og da": -0.293,
    "Av og til": -0.293,
    "Vist": -0.293,
    "Mykje": 0.293,  # nynorsk
    "Utruleg": 0.293,
    "Bestemd": 0.293,
    "Heilt": 0.293,
    "Betydeleg": 0.293,
    "Framfor alt": 0.293,
    "Uvanleg": 0.293,
    "Yttarst": 0.293,
    "Høgt": 0.293,
    "Høgst": 0.293,
    "Særs": 0.293,
    "Serdeles": 0.293,
    "Sers": 0.293,
    "Sjukt": 0.293,
    "Meir": 0.293,
    "Særleg": 0.293,
    "Usadvenleg": 0.293,
    "Utpreget": 0.293,
    "Ekstra": 0.293,
    "Berre": 0.293,
    "Vesentleg": 0.293,
    "Gjennomgåande": 0.293,
    "Sværande": 0.293,
    "Noko": -0.293,
    "So vidt": -0.293,
    "Nokoleis": -0.293,
    "Nokolunde": -0.293,
    "Sjeldan": -0.293,
    "No og da": -0.293,
}

lexicons.register(f"lexicon_{LANG}_v1", func=LEXICON)
lexicons.register(f"negations_{LANG}_v1", func=NEGATIONS)
lexicons.register(f"contrastive_conj_{LANG}_v1", func=CONTRASTIVE_CONJ)
lexicons.register(f"intensifiers_{LANG}_v1", func=INTENSIFIERS)


@Language.factory(f"asent_{LANG}_v1", default_config={"force": True})
def create_no_sentiment_component(nlp: Language, name: str, force: bool) -> Asent:
    f"""Allows the Norwegian sentiment to be added to a spaCy pipe using
    nlp.add_pipe("asent_{LANG}_v1")."""

    LEXICON.update(E_LEXICON)

    return Asent(
        nlp,
        name=name,
        lexicon=LEXICON,
        intensifiers=INTENSIFIERS,
        negations=NEGATIONS,
        contrastive_conjugations=CONTRASTIVE_CONJ,
        lowercase=True,
        lemmatize=False,
        force=force,
    )


components.register(f"asent_{LANG}_v1", func=create_no_sentiment_component)



from typing import Dict, Set

from spacy.language import Language

from ..component import Asent
from ..constants import B_DECR, B_INCR
from ..utils import LEXICON_PATH, components, lexicons, read_lexicon

LANG = "sv"
LEXICON = read_lexicon(LEXICON_PATH / f"{LANG}_lexicon_v1.txt")
NEGATIONS = {
    "inte",
    "varken",
    "aldrig",
    "ingen",
    "nej",
    "ingenting",
    "ingenstans",
    "utan",
    "sällan",
    "trots",
}
CONTRASTIVE_CONJ: Set[str] = set()
INTENSIFIERS: Dict[str, float] = {
    "absolut": B_INCR,
    "otroligt": B_INCR,
    "väldigt": B_INCR,
    "helt": B_INCR,
    "betydande": B_INCR,
    "betydligt": B_INCR,
    "bestämt": B_INCR,
    "djupt": B_INCR,
    "effing": B_INCR,
    "enorm": B_INCR,
    "enormt": B_INCR,
    "framförallt": B_INCR,
    "speciellt": B_INCR,
    "exceptionell": B_INCR,
    "exceptionellt": B_INCR,
    "extrem": B_INCR,
    "extremt": B_INCR,
    "ytterst": B_INCR,
    "fantastiskt": B_INCR,
    "flipping": B_INCR,
    "flippin": B_INCR,
    "frackin": B_INCR,
    "fracking": B_INCR,
    "fricking": B_INCR,
    "frickin": B_INCR,
    "frigging": B_INCR,
    "friggin": B_INCR,
    "fullt": B_INCR,
    "fuckin": B_INCR,
    "fucking": B_INCR,
    "fuggin": B_INCR,
    "fugging": B_INCR,
    "mycket": B_INCR,
    "hella": B_INCR,
    "högt": B_INCR,
    "högst": B_INCR,
    "intensivt": B_INCR,
    "sjukt": B_INCR,
    "mestadels": B_INCR,
    "mer": B_INCR,
    "mest": B_INCR,
    "särskilt": B_INCR,
    "enbart": B_INCR,
    "ganska": B_INCR,
    "anmärkningsvärt": B_INCR,
    "så": B_INCR,
    "väsentligen": B_INCR,
    "genomgående": B_INCR,
    "total": B_INCR,
    "grundligt": B_INCR,
    "totalt": B_INCR,
    "oerhört": B_INCR,
    "uber": B_INCR,
    "ovanligt": B_INCR,
    "nästan": B_DECR,
    "knappt": B_DECR,
    "marginellt": B_DECR,
    "med nöd och näpppe": B_DECR,
    "ungefär": B_DECR,
    "någorlunda": B_DECR,
    "mindre": B_DECR,
    "lite": B_DECR,
    "litet": B_DECR,
    "ibland": B_DECR,
    "stundvis": B_DECR,
    "delvis": B_DECR,
    "sällsynt": B_DECR,
    "då och då": B_DECR,
    "viss": B_DECR,
}

lexicons.register(f"lexicon_{LANG}_v1", func=LEXICON)
lexicons.register(f"negations_{LANG}_v1", func=NEGATIONS)
lexicons.register(f"intensifiers_{LANG}_v1", func=INTENSIFIERS)


@Language.factory(f"asent_{LANG}_v1", default_config={"force": True})
def create_sv_sentiment_component(nlp: Language, name: str, force: bool) -> Asent:
    f"""Allows the Swedish sentiment to be added to a spaCy pipe using
    nlp.add_pipe("asent_{LANG}_v1")."""

    LEXICON.update(LEXICON)

    return Asent(
        nlp,
        name=name,
        lexicon=LEXICON,
        intensifiers=INTENSIFIERS,
        negations=NEGATIONS,
        contrastive_conjugations=set(),
        lowercase=True,
        lemmatize=False,
        force=force,
    )


components.register(f"asent_{LANG}_v1", func=create_sv_sentiment_component)



import math
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

from spacy.tokens import Doc, Span, Token

from .constants import AFTER_BUT_SCALAR, BEFORE_BUT_SCALAR, C_INCR, N_SCALAR
from .data_classes import DocPolarityOutput, SpanPolarityOutput, TokenPolarityOutput


def make_txt_getter(lemmatize: bool, lowercase: bool) -> Callable[[Token], str]:
    """Creates a token getter which return the string of a text string.

    Args:
        lemmatize (bool): Should the token be lemmatized?
        lowercase (bool): Should the token be lowercased?

    Returns:
        Callable[[Token], str]: The getter function
    """

    def get_lower_lemma(token: Token) -> str:
        return getattr(token, "lemma_").lower()

    def get_lemma(token: Token) -> str:
        return getattr(token, "lemma_")

    def get_lower_txt(token: Token) -> str:
        return getattr(token, "text").lower()

    def get_txt(token: Token) -> str:
        return getattr(token, "text")

    if lemmatize:
        if lowercase:
            return get_lower_lemma
        return get_lemma
    if lowercase:
        return get_lower_txt
    return get_txt


def make_intensifier_getter(
    intensifiers: Dict[str, float],
    lemmatize: bool = True,
    lowercase: bool = True,
) -> Callable[[Token], float]:
    """Creates a token getter which checks it token is an intensifier and
    return it intensification factor.

    Args:
        intensifiers (Dict[str, float]): A dictionary of intensifiers and multiplication factor.
        lemmatize (bool): Should it look up in the intensifiers using the lemma? Defaults to True.
        lowercase (bool): Should it look up in the intensifiers using the lowercased word? Defaults to True.

    Returns:
        Callable[[Token], float]: The getter function
    """
    t_getter = make_txt_getter(lemmatize, lowercase)

    if not Doc.has_extension("is_cap_diff"):
        Doc.set_extension("is_cap_diff", getter=allcap_differential_getter)

    def intensifier_scalar_getter(token: Token) -> float:
        """get intensifier score for token."""
        t = t_getter(token)
        if t in intensifiers:
            scalar = intensifiers[t]
            if token.is_upper and token.doc._.is_cap_diff:
                scalar += C_INCR
            return scalar
        return 0.0

    return intensifier_scalar_getter


def allcap_differential_getter(span: Union[Span, Doc]) -> bool:
    """Check whether just some words in the span are ALL CAPS.

    Args:
        span (Union[Span, Doc]): A spaCy span

    Returns:
        bool: `True` if some but not all items in `words` are ALL CAPS
    """
    is_different = False
    allcap_words = 0
    for word in span:
        if word.is_upper:
            allcap_words += 1
    cap_differential = len(span) - allcap_words
    if 0 < cap_differential < len(span):
        is_different = True
    return is_different


def make_valance_getter(
    lexicon: Dict[str, float],
    lemmatize: bool = False,
    lowercase: bool = True,
    cap_differential: float = C_INCR,
) -> Callable[[Token], float]:
    """Creates a token getter which return the valence (sentiment) of a token
    including the capitalization of the token.

    Args:
        lexicon (Dict[str, float]): The valence scores of the tokens.
        lemmatize (bool): Should it look up in the lexicon (and intensifiers)
            using the lemma? Defaults to True.
        lowercase (bool): Should it look up in the lexicon (and intensifiers)
            using the lowercased word? Defaults to True.
        cap_differential (float): Capitalization differential, which is added to the
            valence of the score it is emphasized using all caps. Defaults to 0.733, an
            emperically derived constant (Hutto and Gilbert, 2014). If None it will not
            be used.

    Returns:
        Callable[[Token], float]: The getter function
    """

    t_getter = make_txt_getter(lemmatize, lowercase)

    def valence_getter(token: Token) -> float:
        t = t_getter(token)
        if (t in lexicon) and not (
            Token.has_extension("intensifier") and token._.intensifier
        ):  # if token isn't a intensifier
            return lexicon[t]
        return 0.0

    def cap_diff_valence_getter(token: Token) -> float:
        valence = valence_getter(token)
        if token.is_upper and token.sent._.is_cap_diff:
            if valence > 0:
                valence += cap_differential
            elif valence < 0:
                valence -= cap_differential
        return valence

    if cap_differential:
        if not Span.has_extension("is_cap_diff"):
            Span.set_extension(
                "is_cap_diff",
                getter=allcap_differential_getter,
                force=True,
            )
        return cap_diff_valence_getter
    return valence_getter


def make_is_negation_getter(
    negations: Iterable[str],
    lemmatize: bool = True,
    lowercase: bool = True,
) -> Callable[[Token], bool]:
    """Creates a token getter which return whether a token is a negation or
    not.

    Args:
        negations (Iterable[str]): An list of negations
        lemmatize (bool): Should it look up in negations using the lemma? Defaults to True.
        lowercase (bool): Should it look up in negations using the lowercased word? Defaults to True.

    Returns:
        Callable[[Token], bool]: A token getter
    """
    t_getter = make_txt_getter(lemmatize, lowercase)

    def is_negation(token: Token) -> bool:
        """Determine if token is negation words."""
        t = t_getter(token)
        return t in negations

    return is_negation


def make_is_contrastive_conj_getter(
    contrastive_conjugations: Iterable[str],
    lemmatize: bool = True,
    lowercase: bool = True,
) -> Callable[[Token], bool]:
    """Creates a token getter for whether a token is a constrastive
    conjugations.

    Args:
        contrastive_conjugations (Iterable[str]): A list of contrastive conjugations.
        lemmatize (bool): Should you lemmatize before lookup in the contrastive conjugations list? Defaults to True.
        lowercase (bool): Should you lowercase before lookup in the contrastive conjugations list? Defaults to True.

    Returns:
        Callable[[Token], bool]: A token getter
    """

    t_getter = make_txt_getter(lemmatize, lowercase)

    def is_contrastive_conj(token: Token) -> bool:
        """Determine if token is a contrastive conjugation."""
        t = t_getter(token)
        return t in contrastive_conjugations

    return is_contrastive_conj


def make_is_negated_getter(
    lookback: int = 3,
    is_negation_getter: Optional[Callable[[Token], bool]] = None,
) -> Callable[[Token], Optional[Token]]:
    """Creates a getter which checks if a token is negated by checked whether
    the n previous workds are negations.

    Args:
        lookback (int): How many token should it look backwards for negations? Defaults to 3
            which is emperically derived by Hutto and Gilbert (2014).
        is_negation_getter (Optional[Callable[[Token], bool]]): A function which given a
            token return if the token is a negation or not. If None it assumes that the the token
            extention "is_negation" is set. If specified overwrites teh extension. Defualts to None.

    Returns:
        Callable[[Token], bool]: The getter function
    """
    if is_negation_getter:
        Token.set_extension("is_negation", getter=is_negation_getter, force=True)
    if not Token.has_extension("is_negation"):
        raise ValueError(
            "Token class has no extension 'is_negation', either set the extension"
            + " or provide the is_negation_getter.",
        )

    def is_negated_getter(token: Token) -> Optional[Token]:
        """Determine if token is negated."""
        # don't look back before the sentence start
        min_token_idx = max(token.sent.start, token.i - lookback)
        for t in token.doc[min_token_idx : token.i]:
            if t._.is_negation:
                return t
        return None

    return is_negated_getter


def make_token_polarity_getter(  # noqa: C901
    valence_getter: Optional[Callable[[Token], float]] = None,
    is_negated_getter: Optional[Callable[[Token], Union[bool, Optional[Token]]]] = None,
    intensifier_getter: Optional[Callable[[Token], float]] = None,
    negation_scalar: float = N_SCALAR,
    lookback_intensities: List[float] = [1.0, 0.95, 0.90],
    **kwargs
) -> Callable[[Token], TokenPolarityOutput]:
    """Creates a function (getter) which takes a token and return the polarity
    of the token based upon whether the token valence (sentiment) including
    whether is negated and whether it is intensified using an intensifier.

    Args:
        valence_getter (Optional[Callable[[Token], float]]): a function which given a
            token return the valence (sentiment) of the token. If None it assumes that
            the the token extention "valence" is set. If specified it overwrites the
            extension. Defualts to None.
        is_negated_getter (Optional[Callable[[Token], bool]]): A function which given a
            token return a boolean indicating whether the token is negated. If None it
            assumes that the the token extention "is_negated" is set. If specified it
            overwrites the extension. Defualts to None.
        intensifier_getter (Optional[Callable[[Token], float]]): A getter which for a
            token return 0 if it is not an intensifier or its intensifcation value if it
            is an intensifier. E.g. the token 'especially' might have an value of 0.293
            which increases or decreases the valence of the following word by the
            specified amount. Defaults to None, intensifiers aren't included in the
            analysis of token valence.
        negation_scalar (float):  Defaults to the emperically derived constant
            -0.74 (Hutto and Gilbert, 2014).
        lookback_intensities (List[float]): How long to look back for intensifiers
            (length). Intensities indicate the how much to weight each intensifier.
            Defaults to [1.0, 0.95, 0.90] which is emperically derived
            (Hutto and Gilbert, 2014).
        **kwargs: Additional keyword arguments.

    Returns:
        Callable[[Token], TokenPolarityOutput]: The getter function
    """
    lookback = len(lookback_intensities)

    if valence_getter:
        Token.set_extension("valence", getter=valence_getter, force=True)

    if not Token.has_extension("valence"):
        raise ValueError(
            "Token class has no extension 'valence', either set the extension"
            + " or provide the valence_getter.",
        )
    if is_negated_getter:
        Token.set_extension(
            "is_negated",
            getter=is_negated_getter,
            force=True,
        )
    if not Token.has_extension("is_negated"):
        raise ValueError(
            "Token class has no extension 'is_negated', either set the extension"
            + " or provide the is_negation_getter.",
        )

    if intensifier_getter:
        Token.set_extension("intensifier", getter=intensifier_getter, force=True)
    if not Token.has_extension("intensifier"):
        Token.set_extension(
            "intensifier",
            getter=make_intensifier_getter(
                intensifiers={},
                lemmatize=False,
                lowercase=False,
            ),
        )

    def token_polarity_getter(
        token: Token,
    ) -> TokenPolarityOutput:
        valence = token._.valence

        intensifiers = []
        is_neg = None
        start_tok = token.i  # only used if span is returned
        if valence:
            for start_i in range(1, lookback + 1):
                # dampen the scalar modifier of preceding words and emoticons
                # (excluding the ones that immediately preceed the item) based
                # on their distance from the current item.

                if token.i > start_i:
                    tok_id = token.i - start_i
                    # stop if before sentence start
                    if tok_id < token.sent.start:
                        break

                    prev_token = token.doc[tok_id]
                    b = prev_token._.intensifier
                    if b != 0:
                        intensifiers.append(prev_token)
                        b = b * lookback_intensities[start_i - 1]
                        start_tok = prev_token.i
                    if valence > 0:
                        valence = valence + b
                    else:
                        valence = valence - b
            is_neg = token._.is_negated
            if is_neg:
                valence = valence * negation_scalar
                if isinstance(is_neg, Token) and start_tok > is_neg.i:
                    start_tok = is_neg.i

        return TokenPolarityOutput(
            polarity=valence,
            token=token,
            span=token.doc[start_tok : token.i + 1],
            negation=is_neg,
            intensifiers=intensifiers,
        )

    return token_polarity_getter


def normalize(score: float, alpha=15) -> float:
    """Normalize the score to be between -1 and 1 using an alpha that
    approximates the max expected value.

    Args:
        score (float): The score to normalize.
        alpha (float): The alpha value to use for normalization. Defaults to 15.

    Returns:
        float: The normalized score.
    """
    norm_score = score / math.sqrt((score * score) + alpha)
    if norm_score < -1.0:
        return -1.0
    if norm_score > 1.0:
        return 1.0
    return norm_score


def questionmark_amplification(text: str) -> float:
    """Amplified the questions."""
    # check for added emphasis resulting from question marks (2 or 3+)
    qm_count = text.count("?")
    qm_amplifier = 0.0
    if qm_count > 1:
        if qm_count <= 3:
            # (empirically derived mean sentiment intensity rating increase for
            # question marks)
            qm_amplifier = qm_count * 0.18
        else:
            qm_amplifier = 0.96
    return qm_amplifier


def exclamation_amplification(text: str) -> float:
    # check for added emphasis resulting from exclamation points (up to 4 of them)
    ep_count = text.count("!")
    if ep_count > 4:
        ep_count = 4
    # (empirically derived mean sentiment intensity rating increase for
    # exclamation points)
    ep_amplifier = ep_count * 0.292
    return ep_amplifier


def but_check(
    span: Span,
    sentiment: list,
    before_but_scalar: float = BEFORE_BUT_SCALAR,
    after_but_scalar: float = AFTER_BUT_SCALAR,
) -> list:
    contains_but = False
    for token in span:
        if token._.is_contrastive_conj:
            but_idx = token.i
            contains_but = True
    if contains_but is not False:
        for i, s in enumerate(sentiment[0:but_idx]):
            sentiment[i] = s * before_but_scalar
        for i, s in enumerate(sentiment[but_idx:]):
            sentiment[i] = s * after_but_scalar
    return sentiment


def sift_sentiment_scores(sentiments: Iterable[float]) -> Tuple[float, float, int]:
    """separate positive and negative sentiment scores."""
    pos_sum = 0.0
    neg_sum = 0.0
    neu_count = 0
    for sentiment_score in sentiments:
        if sentiment_score > 0:
            pos_sum += (
                float(sentiment_score) + 1
            )  # compensates for neutral words that are counted as 1
        if sentiment_score < 0:
            neg_sum += (
                float(sentiment_score) - 1
            )  # when used with math.fabs(), compensates for neutrals
        if sentiment_score == 0:
            neu_count += 1
    return pos_sum, neg_sum, neu_count


def make_span_polarity_getter(  # noqa: C901
    polarity_getter: Optional[Callable[[Token], float]],
    contrastive_conj_getter: Optional[Callable[[Token], bool]],
) -> Callable[[Span], SpanPolarityOutput]:
    """Creates a function (getter) which for a span return the aggrated
    polarities. Including accounting for contrastive conjugations (e.g. 'but'),
    exclamationsmarks and questionmarks.

    Args:
        polarity_getter (Optional[Callable[[Token], float]]): A function which given a
            token return the polarity (sentiment) of the token. If None it assumes that
            the token extension "polarity" is already set. If given it overwrites the
            "polarity" extension. Defaults to None.
        contrastive_conj_getter (Optional[Callable[[Token], bool]]): A function which
            given a token return whether the token is a contrastive conjunction. If None
            it assumes that the token extension "is_contrastive_conj" is already set. If
            given it overwrites the "is_contrastive_conj" extension. Defaults to None.

    Returns:
        Callable[[Span], SpanPolarityOutput]: A function which for a span return the
        aggrated polarities. Including accounting for contrastive conjugations (e.g.
        'but'), exclamationsmarks and questionmarks.
    """
    if polarity_getter:
        Token.set_extension("polarity", getter=polarity_getter, force=True)
    if not Token.has_extension("polarity"):
        raise ValueError(
            "Token class has no extension 'polarity', either set the extension"
            + " or provide the polarity_getter.",
        )
    if contrastive_conj_getter:
        Token.set_extension(
            "is_contrastive_conj",
            getter=contrastive_conj_getter,
            force=True,
        )
    if not Token.has_extension("is_contrastive_conj"):
        raise ValueError(
            "Token class has no extension 'is_contrastive_conj', either set the extension"
            + " or provide the contrastive_conj_getter.",
        )

    def __extract(polarity: TokenPolarityOutput) -> Tuple[float, Span]:
        return polarity.polarity, polarity.span

    def __polarity_getter(
        span: Span,
        but_check: Callable = but_check,
    ) -> SpanPolarityOutput:
        polarities = [t._.polarity for t in span]
        sentiment, spans = zip(*[__extract(p) for p in polarities])
        sentiment = but_check(span, list(sentiment))
        sum_s = float(sum(sentiment))

        if sum_s:
            # compute and add emphasis from punctuation in text
            punct_emph_amplifier = 0.0
            punct_emph_amplifier += questionmark_amplification(span.text)
            punct_emph_amplifier += exclamation_amplification(span.text)
            if sum_s > 0:
                sum_s += punct_emph_amplifier
            elif sum_s < 0:
                sum_s -= punct_emph_amplifier

            compound = normalize(sum_s)
            # discriminate between positive, negative and neutral sentiment scores
            pos_sum, neg_sum, neu_count = sift_sentiment_scores(sentiment)

            if pos_sum > math.fabs(neg_sum):
                pos_sum += punct_emph_amplifier
            elif pos_sum < math.fabs(neg_sum):
                neg_sum -= punct_emph_amplifier

            total = pos_sum + math.fabs(neg_sum) + neu_count
            pos = math.fabs(pos_sum / total)
            neg = math.fabs(neg_sum / total)
            neu = math.fabs(neu_count / total)
        else:
            compound = 0.0
            pos = 0.0
            neg = 0.0
            neu = 0.0

        return SpanPolarityOutput(
            negative=neg,
            neutral=neu,
            positive=pos,
            compound=compound,
            span=span,
            polarities=polarities,
        )

    return __polarity_getter


def make_doc_polarity_getter(
    span_polarity_getter: Optional[Callable[[Span], float]],
) -> Callable[[Doc], DocPolarityOutput]:
    """Creates a function (getter) which for a doc return the aggrated
    polarities. Including accounting for contrastive conjugations (e.g. 'but'),
    exclamationsmarks and questionmarks.

    Args:
        span_polarity_getter (Optional[Callable[[Span], float]]):  a function which
            given a span return the polarity (sentiment) of the span. If None it assumes
            that the the span extention "polarity" is already set. If specified it
            overwrites the extension. Defaults to None.

    Returns:
        Callable[[Doc], DocPolarityOutput]: A function which given a doc return the
            aggrated polarities. Including accounting for contrastive conjugations (e.g.
            'but'), exclamationsmarks and questionmarks.
    """

    if span_polarity_getter:
        Span.set_extension("polarity", getter=span_polarity_getter, force=True)
    if not Span.has_extension("polarity"):
        raise ValueError(
            "Span class has no extension 'polarity', either set the extension"
            + " or provide the span_polarity_getter.",
        )

    def polarity_getter(
        doc: Doc,
    ) -> DocPolarityOutput:
        polarities = [sent._.polarity for sent in doc.sents]
        if polarities:
            n = len(polarities)
            negative = sum(p.negative for p in polarities) / n
            positive = sum(p.positive for p in polarities) / n
            neutral = sum(p.neutral for p in polarities) / n
            compound = sum(p.compound for p in polarities) / n
        else:
            negative = 0
            positive = 0
            neutral = 0
            compound = 0

        return DocPolarityOutput(
            negative=negative,
            positive=positive,
            neutral=neutral,
            compound=compound,
            polarities=polarities,
            doc=doc,
        )

    return polarity_getter



from distutils.log import warn
from typing import Union

import spacy
from packaging import version
from spacy import displacy
from spacy.tokens import Doc, Span


def visualize(doc: Union[Span, Doc], style: str = "prediction", cmap="RdYlGn") -> str:
    """Render displaCy visualisation of  model prediction of sentiment or
    analysis of sentiment.

    Args:
        doc (Union[Span, Doc]): The span or document you wish to apply the visualizer to.
        style (str): A string indicating whether it should visualize
            "prediction" or "analysis". "prediction", color codes positive or negative
            spans according to the cmap. "analysis" visualize for each sentimental word
            if it has by negated or intensified a word, and which word.
            If you are looking for the previous visualizer for "prediction", use
            "prediction-no-overlap". Note that this does not allow for overlapping span.
            Thus it can lead to odd results. Defaults to "prediction".
        cmap (str): The color map derived from matplotlib. Defaults to "RdYlGn".

    Returns:
        str: Rendered HTML markup.

    Examples:
        >>> nlp = spacy.load("en_core_web_lg")
        >>> # add the rule-based sentiment model
        >>> nlp.add_pipe("asent_en_v1")
        >>> # try an example
        >>> text = "I am not very happy"
        >>> doc = nlp(text)
        >>> # visualize model prediction
        >>> asent.visualize(doc, style="prediction")
        >>> asent.visualize(doc, style="analysis")
    """

    if style == "prediction" and version.parse(spacy.__version__) < version.parse(
        "3.3.0",
    ):
        warn(
            "The visualization style 'prediction' is not available for spacy version "
            + "< 3.3.0. Using 'prediction-no-overlap' instead. Note that this does not"
            + "allow for overlapping span.",
        )
        style = "prediction-no-overlap"

    if style.lower() == "prediction":
        return visualize_prediction(doc, cmap=cmap)
    elif style.lower() == "prediction-no-overlap":
        return visualize_prediction_no_overlap(doc, cmap=cmap)
    elif style.lower() == "analysis":
        return visualize_analysis(doc)
    else:
        raise ValueError(
            "Invalid style argument, should be either 'analysis' or 'prediction'",
        )


def visualize_prediction_no_overlap(doc: Union[Span, Doc], cmap="RdYlGn") -> str:
    """Render displaCy visualisation of model prediction of sentiment.

    This visualization is similar to visualize_prediction, but it does not allow
    for overlapping spans.

    Args:
        doc (Union[Span, Doc]): The span or document you wish to apply the visualizer to.
        cmap (str, optional): The color map derived from matplotlib. Defaults to "RdYlGn".

    Returns:
        str: Rendered HTML markup.
    """

    if isinstance(doc, Doc):
        span = doc[:]
    else:
        span = doc

    thresholds = [t / 10 for t in range(-50, 51)]
    sentiment_colors = make_colors(n=len(thresholds), cmap=cmap)
    sentiment_color_dict = {str(t): c for c, t in zip(sentiment_colors, thresholds)}

    def __normalize(val: float) -> str:
        return str(max(min(round(val, 1), 5), -5))

    pol = span._.polarity
    t_pols = list(filter(lambda p: p, pol.polarities))

    c_spans = [
        {
            "start": tp.span.doc[tp.span.start].idx - span.doc[span.start].idx,
            "end": tp.span.doc[tp.span.end - 1].idx
            + len(tp.span.doc[tp.span.end - 1].text),
            "label": __normalize(tp.polarity),
        }
        for tp in t_pols
    ]

    ex = [
        {
            "text": span.text,
            "ents": c_spans,
            "title": None,
        },
    ]

    html = displacy.render(
        ex,
        style="ent",
        manual=True,
        options={"colors": sentiment_color_dict},
    )
    return html


def visualize_prediction(doc: Union[Span, Doc], cmap="RdYlGn") -> str:
    """Render displaCy visualisation of model prediction of sentiment.

    Args:
        doc (Union[Span, Doc]): The span or document you wish to apply the visualizer to.
        cmap (str, optional): The color map derived from matplotlib. Defaults to "RdYlGn".

    Returns:
        str: Rendered HTML markup.
    """

    if isinstance(doc, Doc):
        span = doc[:]
    else:
        span = doc

    thresholds = [t / 10 for t in range(-50, 51)]
    sentiment_colors = make_colors(n=len(thresholds), cmap=cmap)
    sentiment_color_dict = {str(t): c for c, t in zip(sentiment_colors, thresholds)}

    def __normalize(val: float) -> str:
        return str(max(min(round(val, 1), 5), -5))

    pol = span._.polarity
    t_pols = list(filter(lambda p: p, pol.polarities))

    c_spans = [
        {
            "start_token": tp.span.start,
            "end_token": tp.span.end,
            "label": __normalize(tp.polarity),
        }
        for tp in t_pols
    ]
    ex = [
        {
            "text": span.text,
            "spans": c_spans,
            "tokens": [t.text for t in span],
        },
    ]

    html = displacy.render(
        ex,
        style="span",
        manual=True,
        options={"colors": sentiment_color_dict},
    )

    return html


def visualize_analysis(doc: Union[Span, Doc]) -> str:
    """Render displaCy visualisation of model analysis.

    Args:
        doc (Union[Span, Doc]): The span or document you wish to apply the visualizer to.

    Returns:
        str: Rendered HTML markup.
    """

    if isinstance(doc, Doc):
        span = doc[:]
    else:
        span = doc

    pol = span._.polarity

    arcs = []
    words = []
    for t, t_pol in zip(span, pol.polarities):
        if t._.valence:
            words.append(
                {
                    "text": t.text,
                    "tag": f"{t._.polarity.polarity:.1f} ({t._.valence:.1f})",
                },
            )
        else:
            words.append({"text": t.text, "tag": f"{t._.valence:.1f}"})

        if t_pol:
            if t_pol.intensifiers:
                for intens in t_pol.intensifiers:
                    arcs.append(
                        {
                            "start": intens.i,
                            "end": t.i,
                            "label": "intensified by",
                            "dir": "left",
                        },
                    )
            if t_pol.negation:
                arcs.append(
                    {
                        "start": t_pol.negation.i,
                        "end": t.i,
                        "label": "negated by",
                        "dir": "left",
                    },
                )
    # Visualize analysis
    from spacy import displacy

    ex = {"words": words, "arcs": arcs}
    html = displacy.render(ex, style="dep", manual=True)

    return html


def make_colors(n=10, cmap="RdYlGn"):
    """A utility function for creating a stepped color gradient."""
    from pylab import cm, matplotlib

    cmap = cm.get_cmap(cmap, n)  # PiYG

    for i in range(cmap.N):
        rgba = cmap(i)
        # rgb2hex accepts rgb or rgba
        yield matplotlib.colors.rgb2hex(rgba)


# def print_colors(HEX: Iterable) -> None:
#     """An utility function for visualizing a color map"""
#     from IPython.core.display import HTML, display

#     for color in HEX:
#         display(HTML(f'<p style="color:{color}">{color}</p>'))



import asent


def test_lexicons():
    d = asent.lexicons.get("emoji_v1")
    assert isinstance(d, dict)

    for lexicon in asent.lexicons.get_all():
        print(lexicon)

    asent.register_lexicon("my_custom_lexicon", {"happy": 2, "sad": -2})
    lex = asent.lexicons.get("my_custom_lexicon")
    assert isinstance(lex, dict)


def test_components():
    f = asent.components.get("asent_da_v1")
    assert callable(f)

    for i, c in asent.components.get_all().items():
        assert callable(c)

    asent.register_component("my_custom_comp", lambda doc: doc)
    comp = asent.components.get("my_custom_comp")
    assert callable(comp)



"""Test specifically targeted an bugs."""

import spacy

import asent  # noqa


def test_no_negations_and_intensifiers_out_of_sentence():
    """Test that no negations are not found outside the sentence span.

    https://github.com/KennethEnevoldsen/asent/issues/58
    """

    # create spacy pipeline
    nlp = spacy.blank("en")
    nlp.add_pipe("sentencizer")

    nlp.add_pipe("asent_en_v1")

    text = "Would you do that? I would not. Very stupid is what that is."
    doc = nlp(text)
    assert doc[10]._.is_negated is None

    text = "Would you do that? I would not very. Stupid is what that is."
    doc = nlp(text)
    assert doc[10]._.valence == doc[10]._.polarity.polarity



import pytest
import spacy

from asent.data_classes import (
    DocPolarityOutput,
    SpanPolarityOutput,
    TokenPolarityOutput,
)


@pytest.fixture()
def sample_doc():
    nlp = spacy.blank("en")
    doc = nlp("sample documents")
    return doc


def test_TokenPolarityOutput(sample_doc):
    t1 = TokenPolarityOutput(polarity=1, token=sample_doc[0], span=sample_doc[:])
    t2 = TokenPolarityOutput(polarity=2, token=sample_doc[0], span=sample_doc[:])

    print(t1)
    assert t1 != t2
    assert t1 < t2
    assert t2 > t1


def test_SpanPolarityOutput(sample_doc):
    tok = TokenPolarityOutput(polarity=1, token=sample_doc[0], span=sample_doc[:])
    t1 = SpanPolarityOutput(
        negative=0,
        neutral=0,
        positive=0,
        compound=1,
        span=sample_doc[:],
        polarities=[tok],
    )
    t2 = SpanPolarityOutput(
        negative=0,
        neutral=0,
        positive=0,
        compound=2,
        span=sample_doc[:],
        polarities=[tok],
    )

    print(t1)
    assert t1 != t2
    assert t1 < t2
    assert t2 > t1


def test_DocPolarityOutput(sample_doc):
    tok = TokenPolarityOutput(polarity=1, token=sample_doc[0], span=sample_doc[:])
    span = SpanPolarityOutput(
        negative=0,
        neutral=0,
        positive=0,
        compound=1,
        span=sample_doc[:],
        polarities=[tok],
    )

    t1 = DocPolarityOutput(
        negative=0,
        neutral=0,
        positive=0,
        compound=1,
        span=sample_doc[:],
        polarities=[span],
        doc=sample_doc,
    )
    t2 = DocPolarityOutput(
        negative=0,
        neutral=0,
        positive=0,
        compound=2,
        span=sample_doc[:],
        polarities=[span],
        doc=sample_doc,
    )

    print(t1)
    assert t1 != t2
    assert t1 < t2
    assert t2 > t1



from functools import partial
from warnings import warn

from spacy.language import Language

from asent.component import Asent
from asent.lang.emoji import LEXICON as E_LEXICON
from asent.utils import LEXICON_PATH, components, lexicons, read_lexicon

lang_path = LEXICON_PATH / ".." / "lang"
langs = [lang.stem for lang in lang_path.glob("*.py") if len(lang.stem) < 4]


def create_xx_sentiment_component(
    nlp: Language,
    name: str,
    lang: str,
    force: bool,
) -> Asent:
    """Allows the sentiment pipe to be added to a spaCy pipe using
    nlp.add_pipe("asent_{language id}_v1"). Note that this is overwritten by
    languages which have a specified default, e.g. the case for en, da, sv.

    note that this function uses a lexicon that is automatically
    constructed. We therefore recommend examines the output and
    adjusting the lexicon. For more information on how the lexicon was
    constructed see; Chen, Y., & Skiena, S. (2014). Building Sentiment
    Lexicons for All Major Languages.

    Args:
        nlp: The spaCy language object to add the sentiment pipe to.
        name: The name of the sentiment pipe.
        lang: The language id of the sentiment pipe.
        force: Whether to force the sentiment pipe to be added to the language.

    Returns:
        Asent: The sentiment pipe.
    """

    msg = (
        f"'asent_{lang}_v1' uses a lexicon that is automatically constructed. We "
        + "therefore recommend examines the output and adjusting the lexicon. "
        + "For more information on how the lexicon was constructed see; Chen, Y., & "
        + "Skiena, S. (2014). Building Sentiment Lexicons for All Major Languages."
    )
    warn(msg)

    lex = lexicons.get(f"lexicon_{lang}_chen_skiena_2014_v1.txt")
    lex.update(E_LEXICON)

    return Asent(
        nlp,
        name=name,
        lexicon=lex,
        intensifiers={},
        negations={},
        contrastive_conjugations={},
        lowercase=True,
        lemmatize=False,
        force=force,
    )


_create_xx_component = {}

for lex_path in LEXICON_PATH.glob("*_lexicon_chen_skiena_2014_v1.txt"):
    lang = lex_path.stem.split("_")[0]

    # register the Chen Skiena (2014) lexicons
    lexicons.register(
        f"lexicon_{lang}_chen_skiena_2014_v1.txt",
        func=read_lexicon(LEXICON_PATH / f"{lang}_lexicon_chen_skiena_2014_v1.txt"),
    )

    # if there is no specified default for the language specify the default to be the
    # autogenerated lexicon.
    if lang in langs:
        continue
    lexicons.register(
        f"lexicon_{lang}_v1",
        func=lexicons.get(f"lexicon_{lang}_chen_skiena_2014_v1.txt"),
    )
    lex = lexicons.get(f"lexicon_{lang}_v1")
    lex.update(E_LEXICON)

    _create_xx_component[lang] = partial(create_xx_sentiment_component, lang=lang)

    Language.factory(
        f"asent_{lang}_v1",
        default_config={"force": True},
        func=_create_xx_component[lang],
    )

    components.register(f"asent_{lang}_v1", func=_create_xx_component[lang])



from pathlib import Path
from typing import Callable, Dict, Union

import catalogue

LEXICON_PATH = Path(__file__).parent / ".." / "asent" / "lexicons"

lexicons = catalogue.create("asent", "lexicon", entry_points=True)
components = catalogue.create("asent", "components", entry_points=True)


def register_lexicon(name: str, lexicon: Dict[str, float]) -> None:
    """Registers a lexicon in asent.lexicons.

    Args:
        name (str): The name of the lexicon
        lexicon (Dict[str, float]): The lexicon supplies as a dictionary.

    Example:
        >>> asent.register("my_lexicon_v1", {"happy": 4, "sad": -2})
        >>> asent.lexicons.get("my_lexicon_v1")
        {"happy": 4, "sad": -2}
    """
    lexicons.register(name, func=lexicon)


def register_component(name: str, func: Callable) -> None:
    """Registers a component in asent.components.

    Args:
        name (str): The name of the lexicon
        func (Callable): A Callable component
    """
    components.register(name, func=func)


def read_lexicon(path: Union[str, Path]) -> Dict[str, float]:
    with open(path, encoding="utf-8") as f:
        lexicon = {}
        for line in f.read().rstrip("\n").split("\n"):
            if not line:
                continue
            (word, measure) = line.strip().split("\t")[0:2]
            lexicon[word] = float(measure)
    return lexicon



from .about import __documentation__, __download_url__, __title__, __version__  # noqa
from .lang import da, emoji, en, no, sv  # noqa
from .utils import components, lexicons, register_component, register_lexicon  # noqa
from .visualize import visualize  # noqa



from spacy.language import Language

from ..component import Asent
from ..constants import B_INCR
from ..utils import LEXICON_PATH, components, lexicons, read_lexicon
from .emoji import LEXICON as E_LEXICON

LANG = "da"


def read_csv_lexicon():
    lexicon_file = LEXICON_PATH / f"{LANG}_lexicon_sentida_lemma_v1.csv"

    with open(lexicon_file, encoding="utf-8") as f:
        pairs = filter(lambda x: x, f.read().split("\n")[1:])
        lexicon = {
            word: float(rating) for word, rating in map(lambda x: x.split(","), pairs)
        }
    return lexicon


LEXICON = read_csv_lexicon()
NEGATIONS = {"ikke", "ik", "ikk", "ik'", "aldrig", "ingen"}
CONTRASTIVE_CONJ = {"men", "dog"}
INTENSIFIERS = {
    "absolut": B_INCR,
    "vældig": B_INCR,
    "helt": B_INCR,
    "betydende": B_INCR,
    "betydelig": B_INCR,
    "bestemt": B_INCR,
    "enorm": B_INCR,
    "exceptionel": B_INCR,
    "ekseptionel": B_INCR,
    "extrem": B_INCR,
    "yderst": B_INCR,
    "fantastisk": B_INCR,
    "flipping": B_INCR,
    "flippin": B_INCR,
    "frackin": B_INCR,
    "fracking": B_INCR,
    "fricking": B_INCR,
    "frickin": B_INCR,
    "frigging": B_INCR,
    "friggin": B_INCR,
    "fuckin": B_INCR,
    "fucking": B_INCR,
    "fuggin": B_INCR,
    "fugging": B_INCR,
    "hella": B_INCR,
    "intensiv": B_INCR,
    "mest": B_INCR,
    "særskilt": B_INCR,
    "ganske": B_INCR,
    "væsentlig": B_INCR,
    "uber": B_INCR,
    "virkelig": B_INCR,
    "særlig": B_INCR,
    "særligt": B_INCR,
    "temmelig": 0.1,
    "megen": 0.2,
    "mega": 0.4,
    "lidt": -0.2,
    "ekstrem": 0.4,
    "total": 0.2,
    "utrolig": 0.3,
    "rimelig": 0.1,
    "seriøs": 0.3,
}


lexicons.register(f"lexicon_{LANG}_sentida_lemma_v1", func=LEXICON)
lexicons.register(f"negations_{LANG}_v1", func=NEGATIONS)
lexicons.register(f"contrastive_conj_{LANG}_v1", func=CONTRASTIVE_CONJ)
lexicons.register(f"intensifiers_{LANG}_v1", func=INTENSIFIERS)

lexicons.register(
    "lexicon_da_afinn_v1",
    func=read_lexicon(LEXICON_PATH / f"{LANG}_lexicon_afinn_v1.txt"),
)
lexicons.register(  # store as default
    f"lexicon_{LANG}_v1",
    func=lexicons.get(f"lexicon_{LANG}_afinn_v1"),
)


@Language.factory(f"asent_{LANG}_v1", default_config={"force": True})
def create_da_sentiment_component(nlp: Language, name: str, force: bool) -> Asent:
    f"""Allows the Danish sentiment to be added to a spaCy pipe using
    nlp.add_pipe("asent_{LANG}_v1")."""

    lex = lexicons.get(f"lexicon_{LANG}_afinn_v1")
    lex.update(E_LEXICON)

    return Asent(
        nlp,
        name=name,
        lexicon=lex,
        intensifiers=INTENSIFIERS,
        negations=NEGATIONS,
        contrastive_conjugations=CONTRASTIVE_CONJ,
        lowercase=True,
        lemmatize=False,
        force=force,
    )


components.register(f"asent_{LANG}_v1", func=create_da_sentiment_component)



from spacy.language import Language

from ..component import Asent
from ..constants import B_DECR, B_INCR
from ..utils import LEXICON_PATH, components, lexicons, read_lexicon
from .emoji import LEXICON as E_LEXICON

LANG = "en"
LEXICON = read_lexicon(LEXICON_PATH / f"{LANG}_lexicon_v1.txt")
NEGATIONS = {
    "aint",
    "arent",
    "cannot",
    "cant",
    "couldnt",
    "darent",
    "didnt",
    "doesnt",
    "ain't",
    "aren't",
    "can't",
    "couldn't",
    "daren't",
    "didn't",
    "doesn't",
    "dont",
    "hadnt",
    "hasnt",
    "havent",
    "isnt",
    "mightnt",
    "mustnt",
    "neither",
    "don't",
    "hadn't",
    "hasn't",
    "haven't",
    "isn't",
    "mightn't",
    "mustn't",
    "neednt",
    "needn't",
    "never",
    "none",
    "nope",
    "nor",
    "not",
    "nothing",
    "nowhere",
    "oughtnt",
    "shant",
    "shouldnt",
    "uhuh",
    "wasnt",
    "werent",
    "oughtn't",
    "shan't",
    "shouldn't",
    "uh-uh",
    "wasn't",
    "weren't",
    "without",
    "wont",
    "wouldnt",
    "won't",
    "wouldn't",
    "rarely",
    "seldom",
    "despite",
}
CONTRASTIVE_CONJ = {"but"}
INTENSIFIERS = {
    "absolutely": B_INCR,
    "amazingly": B_INCR,
    "awfully": B_INCR,
    "completely": B_INCR,
    "considerable": B_INCR,
    "considerably": B_INCR,
    "decidedly": B_INCR,
    "deeply": B_INCR,
    "effing": B_INCR,
    "enormous": B_INCR,
    "enormously": B_INCR,
    "entirely": B_INCR,
    "especially": B_INCR,
    "exceptional": B_INCR,
    "exceptionally": B_INCR,
    "extreme": B_INCR,
    "extremely": B_INCR,
    "fabulously": B_INCR,
    "flipping": B_INCR,
    "flippin": B_INCR,
    "frackin": B_INCR,
    "fracking": B_INCR,
    "fricking": B_INCR,
    "frickin": B_INCR,
    "frigging": B_INCR,
    "friggin": B_INCR,
    "fully": B_INCR,
    "fuckin": B_INCR,
    "fucking": B_INCR,
    "fuggin": B_INCR,
    "fugging": B_INCR,
    "greatly": B_INCR,
    "hella": B_INCR,
    "highly": B_INCR,
    "hugely": B_INCR,
    "incredible": B_INCR,
    "incredibly": B_INCR,
    "intensely": B_INCR,
    "major": B_INCR,
    "majorly": B_INCR,
    "more": B_INCR,
    "most": B_INCR,
    "particularly": B_INCR,
    "purely": B_INCR,
    "quite": B_INCR,
    "really": B_INCR,
    "remarkably": B_INCR,
    "so": B_INCR,
    "substantially": B_INCR,
    "thoroughly": B_INCR,
    "total": B_INCR,
    "totally": B_INCR,
    "tremendous": B_INCR,
    "tremendously": B_INCR,
    "uber": B_INCR,
    "unbelievably": B_INCR,
    "unusually": B_INCR,
    "utter": B_INCR,
    "utterly": B_INCR,
    "very": B_INCR,
    "almost": B_DECR,
    "barely": B_DECR,
    "hardly": B_DECR,
    "just enough": B_DECR,
    "kind of": B_DECR,
    "kinda": B_DECR,
    "kindof": B_DECR,
    "kind-of": B_DECR,
    "less": B_DECR,
    "little": B_DECR,
    "marginal": B_DECR,
    "marginally": B_DECR,
    "occasional": B_DECR,
    "occasionally": B_DECR,
    "partly": B_DECR,
    "scarce": B_DECR,
    "scarcely": B_DECR,
    "slight": B_DECR,
    "slightly": B_DECR,
    "somewhat": B_DECR,
    "sort of": B_DECR,
    "sorta": B_DECR,
    "sortof": B_DECR,
    "sort-of": B_DECR,
}

lexicons.register(f"lexicon_{LANG}_v1", func=LEXICON)
lexicons.register(f"negations_{LANG}_v1", func=NEGATIONS)
lexicons.register(f"contrastive_conj_{LANG}_v1", func=CONTRASTIVE_CONJ)
lexicons.register(f"intensifiers_{LANG}_v1", func=INTENSIFIERS)


@Language.factory(f"asent_{LANG}_v1", default_config={"force": True})
def create_en_sentiment_component(nlp: Language, name: str, force: bool) -> Asent:
    f"""Allows the English sentiment to be added to a spaCy pipe using
    nlp.add_pipe("asent_{LANG}_v1")."""
    LEXICON.update(E_LEXICON)

    return Asent(
        nlp,
        name=name,
        lexicon=LEXICON,
        intensifiers=INTENSIFIERS,
        negations=NEGATIONS,
        contrastive_conjugations=CONTRASTIVE_CONJ,
        lowercase=True,
        lemmatize=False,
        force=force,
    )


components.register(f"asent_{LANG}_v1", func=create_en_sentiment_component)



import spacy

import asent


def test_all_lang_components():
    """Test that all language components can be loaded and run on a simple text
    snippet."""
    text = "this is a random test string"
    nlp = spacy.blank("xx")  # multilingual pipe

    for comp in asent.components.get_all():
        nlp.add_pipe(comp, config={"force": True})
        nlp.remove_pipe(comp)
        doc = nlp(text)  # noqa



import pytest
import spacy
from spacy.tokens import Doc, Token

import asent
from asent.getters import (
    make_doc_polarity_getter,
    make_intensifier_getter,
    make_is_contrastive_conj_getter,
    make_is_negated_getter,
    make_is_negation_getter,
    make_span_polarity_getter,
    make_token_polarity_getter,
    make_valance_getter,
)


@pytest.fixture()
def nlp_dict():
    nlp_en = spacy.blank("en")
    nlp_en.add_pipe("sentencizer")
    nlp_da = spacy.load("da_core_news_sm")
    nlp_sv = spacy.blank("sv")
    nlp_sv.add_pipe("sentencizer")
    nlp_nb = spacy.blank("nb")
    nlp_nb.add_pipe("sentencizer")
    nlp_de = spacy.blank("de")
    nlp_de.add_pipe("sentencizer")

    return {"da": nlp_da, "en": nlp_en, "no": nlp_nb, "sv": nlp_sv, "de": nlp_de}


@pytest.mark.parametrize(
    "example,idx,expected,lang",
    [
        ("jeg er glad", -1, "positive", "da"),
        ("jeg er GLAD", -1, "positive", "da"),
        ("jeg er sur", -1, "negative", "da"),
    ],
)
def test_valence_getter(example, idx, expected, lang: str, nlp_dict):
    nlp = nlp_dict[lang]
    lexicon = asent.lexicons.get("lexicon_" + lang + "_v1")

    Token.set_extension(
        "valence",
        getter=make_valance_getter(lexicon=lexicon),
        force=True,
    )
    doc = nlp(example)

    if expected == "positive":
        assert doc[idx]._.valence > 0
    elif expected == "negative":
        assert doc[idx]._.valence < 0
    elif expected == "neutral":
        assert doc[idx]._.valence == 0
    else:
        raise ValueError(
            "Invalid expected value '{expected}', should be either 'positive' or 'negative'",
        )


@pytest.mark.parametrize(
    "example,cased_example,idx,lang",
    [
        ("jeg er glad", "jeg er GLAD", -1, "da"),
        ("jeg er sur", "jeg er SUR", -1, "da"),
    ],
)
def test_valence_getter_casing(example, cased_example, idx, lang, nlp_dict):
    nlp = nlp_dict[lang]
    lexicon = asent.lexicons.get("lexicon_" + lang + "_v1")

    Token.set_extension(
        "valence",
        getter=make_valance_getter(lexicon=lexicon),
        force=True,
    )

    docs = [nlp(t) for t in [example, cased_example]]
    assert abs(docs[0][idx]._.valence) < abs(docs[1][idx]._.valence)


@pytest.mark.parametrize(
    "example,idx,is_negated,lang",
    [
        ("jeg er glad", -1, False, "da"),
        ("jeg er GLAD", -1, False, "da"),
        ("jeg er sur", -1, False, "da"),
        ("jeg er ikke sur", -1, True, "da"),
        ("jeg er ikke længere sur", -1, True, "da"),
        ("jeg er ikke længere særligt sur", -1, True, "da"),
    ],
)
def test_is_negation(example, idx, is_negated, lang: str, nlp_dict):
    nlp = nlp_dict[lang]
    negations = asent.lexicons.get("negations_" + lang + "_v1")

    is_negation_getter = make_is_negation_getter(negations=negations)
    is_negated_getter = make_is_negated_getter(is_negation_getter=is_negation_getter)
    Token.set_extension(name="is_negated", getter=is_negated_getter, force=True)
    doc = nlp(example)
    assert bool(doc[idx]._.is_negated) is is_negated


@pytest.mark.parametrize(
    "example,idx,more_positive_example,more_positive_idx,lang",
    [
        ("jeg er glad", -1, "jeg er GLAD", -1, "da"),
        ("jeg er ikke glad", -1, "jeg er glad", -1, "da"),
    ],
)
def test_token_polarity(
    example: str,
    idx: int,
    more_positive_example: str,
    more_positive_idx: int,
    lang: str,
    nlp_dict,
):

    nlp = nlp_dict[lang]
    lexicon = asent.lexicons.get("lexicon_" + lang + "_v1")
    negations = asent.lexicons.get("negations_" + lang + "_v1")
    intensifiers = asent.lexicons.get("intensifiers_" + lang + "_v1")

    lowercase = True
    lemmatize = False

    valence_getter = make_valance_getter(
        lexicon=lexicon,
        lowercase=lowercase,
        lemmatize=lemmatize,
    )
    is_negation_getter = make_is_negation_getter(negations=negations)
    is_negated_getter = make_is_negated_getter(is_negation_getter=is_negation_getter)
    intensifier_getter = make_intensifier_getter(
        intensifiers=intensifiers,
        lowercase=lowercase,
        lemmatize=lemmatize,
    )

    polarity_getter = make_token_polarity_getter(
        valence_getter=valence_getter,
        is_negated_getter=is_negated_getter,
        intensifier_getter=intensifier_getter,
    )
    Token.set_extension("polarity", getter=polarity_getter, force=True)

    docs = [nlp(e) for e in [example, more_positive_example]]
    assert docs[0][idx]._.polarity < docs[1][more_positive_idx]._.polarity


@pytest.mark.parametrize(
    "example,expected,lang",
    [
        ("jeg er glad", "positive", "da"),
        ("jeg er sur", "negative", "da"),
        ("jeg er sur?", "negative", "da"),
        ("jeg er sur???", "negative", "da"),
        ("jeg er sur!", "negative", "da"),
        ("jeg er sur!!!!!!", "negative", "da"),
        ("xyz", "neutral", "da"),
        ("filmen var okay god men er generelt skuffet", "negative", "da"),
        ("ich bin ser gut!", "positive", "de"),
    ],
)
def test_span_doc_polarity(example: str, expected: str, lang: str, nlp_dict):
    nlp = nlp_dict[lang]
    lexicon = asent.lexicons.get("lexicon_" + lang + "_v1")

    if lang in {"da", "sv"}:
        negations = asent.lexicons.get("negations_" + lang + "_v1")
        intensifiers = asent.lexicons.get("intensifiers_" + lang + "_v1")
    else:
        negations = set()
        intensifiers = {}
    if lang in {"da"}:
        cconj = asent.lexicons.get("contrastive_conj_" + lang + "_v1")
    else:
        cconj = set()

    lowercase = True
    lemmatize = False

    valence_getter = make_valance_getter(
        lexicon,
        lowercase=lowercase,
        lemmatize=lemmatize,
    )
    is_negation_getter = make_is_negation_getter(negations)
    is_negated_getter = make_is_negated_getter(is_negation_getter=is_negation_getter)
    intensifier_getter = make_intensifier_getter(
        intensifiers,
        lowercase=lowercase,
        lemmatize=lemmatize,
    )
    polarity_getter = make_token_polarity_getter(
        valence_getter=valence_getter,
        is_negated_getter=is_negated_getter,
        intensifier_getter=intensifier_getter,
    )
    contrastive_conj_getter = make_is_contrastive_conj_getter(cconj)
    span_polarity_getter = make_span_polarity_getter(
        polarity_getter,
        contrastive_conj_getter=contrastive_conj_getter,
    )
    doc_polarity_getter = make_doc_polarity_getter(span_polarity_getter)
    Doc.set_extension("polarity", getter=doc_polarity_getter, force=True)
    doc = nlp(example)
    sent = [s for s in doc.sents][0]  # assuming there is only one sentence
    assert doc._.polarity.compound == sent._.polarity.compound
    if expected == "positive":
        assert sent._.polarity.compound > 0
    elif expected == "negative":
        assert sent._.polarity.compound < 0
    elif expected == "neutral":
        assert sent._.polarity.compound == 0
    else:
        raise ValueError(
            "Invalid expected value '{expected}', should be either 'neutral', 'positive' or 'negative'",
        )


@pytest.mark.parametrize(
    "example,more_positive_example,lang",
    [
        ("jeg er sur", "jeg er glad", "da"),
        ("jeg er SUR", "jeg er sur", "da"),
        (
            "filmen var okay god men er generelt skuffet",
            "filmen var okay god, er generelt skuffet",
            "da",
        ),
    ],
)
def test_span_polarity_contrast(
    example: str,
    more_positive_example: str,
    lang: str,
    nlp_dict,
):
    nlp = nlp_dict[lang]

    nlp.add_pipe("asent_" + lang + "_v1", config={"force": True})

    docs = [nlp(e) for e in [example, more_positive_example]]
    assert docs[0]._.polarity.compound < docs[1]._.polarity.compound


@pytest.mark.parametrize(
    "example,expected,lang",
    [
        ("I am happy", "positive", "en"),
        ("I am very happy", "positive", "en"),
        ("I am VERY happy", "positive", "en"),
        ("I am not very happy", "negative", "en"),
        ("jeg er glad", "positive", "da"),
        ("jeg er glad", "positive", "no"),
        ("jag är glad", "positive", "sv"),
        ("", "neutral", "da"),
    ],
)
def test_components(example: str, expected: str, lang: str, nlp_dict):
    nlp = nlp_dict[lang]

    nlp.add_pipe("asent_" + lang + "_v1", config={"force": True})
    doc = nlp(example)
    if expected == "positive":
        assert doc._.polarity.compound > 0
    elif expected == "neutral":
        assert doc._.polarity.compound == 0
    elif expected == "negative":
        assert doc._.polarity.compound < 0
    else:
        raise ValueError(
            "Invalid expected value '{expected}', should be either 'neutral', 'positive' or 'negative'",
        )



import setuptools

with open("asent/about.py") as f:
    v = f.read()
    for line in v.split("\n"):
        if line.startswith("__version__"):
            __version__ = line.split('"')[-2]


with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(version=__version__)



"""This contains emperically derived constants from Hutto and Gilbert (2014)"""

# (empirically derived mean sentiment intensity rating increase for booster words)
B_INCR = 0.293
B_DECR = -0.293

# (empirically derived mean sentiment intensity rating increase for using ALLCAPs to emphasize a word)
C_INCR = 0.733  # capitatilization scaler
N_SCALAR = -0.74


BEFORE_BUT_SCALAR = 0.5
AFTER_BUT_SCALAR = 1.5



from .autoconstructed import create_xx_sentiment_component  # noqa
from .da import create_da_sentiment_component  # noqa
from .emoji import LEXICON  # noqa
from .en import create_en_sentiment_component  # noqa
from .no import create_no_sentiment_component  # noqa
from .sv import create_sv_sentiment_component  # noqa



__title__ = "asent"

__version__ = "0.6.1"  # the ONLY source of version ID
__download_url__ = "https://github.com/kennethenevoldsen/asent"
__documentation__ = "https://kennethenevoldsen.github.io/asent"
